@@ -68,17 +68,15 @@ extern "C" {
 #include "srl_reader_varint.h"
 #include "srl_reader_decompress.h"
 
-#define srl_stack_push_and_set(_iter, _tag, _length) STMT_START {  \
-    srl_iterator_stack_ptr _stack_ptr;                          \
-    srl_stack_push_ptr((_iter)->stack, _stack_ptr);             \
-    _stack_ptr->offset = SRL_RDR_BODY_POS_OFS((_iter)->pbuf);   \
-    _stack_ptr->count = (_length);                              \
-    _stack_ptr->idx = (_length);                                \
-    _stack_ptr->tag = (_tag);                                   \
+#define SRL_RDR_BODY_POS_OFS_(buf) ((buf).pos - (buf).body_pos)
+#define srl_stack_push_and_set(_iter, _tag, _length, _stack_ptr) STMT_START {   \
+    srl_stack_push_ptr((_iter)->pstack, (_stack_ptr));                          \
+    (_stack_ptr)->offset = SRL_RDR_BODY_POS_OFS_((_iter)->buf);                 \
+    (_stack_ptr)->count = (_length);                                            \
+    (_stack_ptr)->idx = (_length);                                              \
+    (_stack_ptr)->tag = (_tag);                                                 \
 } STMT_END
 
-#define srl_stack_type_t srl_iterator_stack_t
-#include "srl_stack.h"
 #define SRL_ITER_STACK_ROOT_TAG SRL_HDR_PACKET_START
 #define SRL_ITER_STACK_ON_ROOT(stack) ((stack)->ptr->tag == SRL_ITER_STACK_ROOT_TAG)
 
@@ -104,12 +102,12 @@ extern "C" {
         );                                                                          \
     } STMT_END
 #   define SRL_ITER_REPORT_STACK_STATE(iter) STMT_START {                           \
-        srl_iterator_stack_ptr stack_ptr = (iter)->stack->ptr;                      \
+        srl_iterator_stack_ptr stack_ptr = (iter)->pstack->ptr;                     \
         SRL_RDR_TRACE(                                                              \
             "%s stack state depth=%"IVdf" tag=SRL_HDR_%s (int: %d hex: 0x%x) "      \
             "count=%u idx=%d offset=%"UVuf,                                         \
             srl_debug_tabulator((iter)),                                            \
-            SRL_STACK_DEPTH((iter)->stack),                                         \
+            SRL_STACK_DEPTH((iter)->pstack),                                        \
             SRL_TAG_NAME(stack_ptr->tag), stack_ptr->tag, stack_ptr->tag,           \
             stack_ptr->count,                                                       \
             stack_ptr->idx,                                                         \
@@ -129,14 +127,14 @@ extern "C" {
 } STMT_END
 
 #define SRL_ITER_ASSERT_STACK(iter) STMT_START {                                    \
-    assert(!srl_stack_empty((iter)->stack));                                        \
-    if (expect_false(((int) (iter)->stack->ptr->idx)) < 0) {                        \
+    assert(!srl_stack_empty((iter)->pstack));                                       \
+    if (expect_false(((int) (iter)->stack.ptr->idx)) < 0) {                         \
         SRL_ITER_ERROR("Stack is empty! Inconsistent state!");                      \
     }                                                                               \
 } STMT_END
 
 #define SRL_ITER_ASSERT_ARRAY_ON_STACK(iter) STMT_START {                           \
-    U8 _tag = iter->stack->ptr->tag;                                                \
+    U8 _tag = iter->stack.ptr->tag;                                                 \
     if (expect_false(                                                               \
             _tag != SRL_HDR_ARRAY                                                   \
          && (_tag < SRL_HDR_ARRAYREF_LOW || _tag > SRL_HDR_ARRAYREF_HIGH)           \
@@ -148,7 +146,7 @@ extern "C" {
 } STMT_END
 
 #define SRL_ITER_ASSERT_HASH_ON_STACK(iter) STMT_START {                            \
-    U8 _tag = iter->stack->ptr->tag;                                                \
+    U8 _tag = iter->stack.ptr->tag;                                                 \
     if (expect_false(                                                               \
             _tag != SRL_HDR_HASH                                                    \
          && (_tag < SRL_HDR_HASHREF_LOW || _tag > SRL_HDR_HASHREF_HIGH)             \
@@ -171,7 +169,7 @@ srl_debug_tabulator(pTHX_ srl_iterator_t *iter)
 {
     int i = 0;
     static char buf[1024];
-    for (; i < SRL_STACK_DEPTH(iter->stack); ++i) buf[i] = '+';
+    for (; i < SRL_STACK_DEPTH(iter->pstack); ++i) buf[i] = '+';
     buf[i] = '\0';
     return (const char *) buf; // XXX
 }
@@ -179,32 +177,27 @@ srl_debug_tabulator(pTHX_ srl_iterator_t *iter)
 srl_iterator_t *
 srl_build_iterator_struct(pTHX_ HV *opt)
 {
-    srl_stack_t *stack = NULL;
     srl_iterator_t *iter = NULL;
-
     Newx(iter, 1, srl_iterator_t);
-    if (iter == NULL)
-        croak("Out of memory");
+    if (iter == NULL) croak("Out of memory");
+    srl_init_iterator(aTHX_ iter, opt);
+    return iter;
+}
 
-    Newx(stack, 1, srl_stack_t);
-    if (stack == NULL) {
-        Safefree(iter);
-        croak("Out of memory");
-    }
+void
+srl_init_iterator(pTHX_ srl_iterator_t *iter, HV *opt)
+{
+    assert(iter != NULL);
 
-    // TODO inline stack
-    // TODO keep fixed stack size ???
-    if (expect_false(srl_stack_init(aTHX_ stack, 32) != 0)) {
+    if (expect_false(srl_stack_init(aTHX_ &iter->stack, 32) != 0)) {
         Safefree(iter);
-        Safefree(stack);
         croak("Out of memory");
     }
 
     SRL_RDR_CLEAR(&iter->buf);
     iter->pbuf = &iter->buf;
-    iter->stack = stack;
+    iter->pstack = &iter->stack;
     iter->document = NULL;
-    iter->tmp_buf_owner = NULL;
     iter->dec = NULL;
 
     /* load options */
@@ -213,46 +206,62 @@ srl_build_iterator_struct(pTHX_ HV *opt)
         if (svp && SvTRUE(*svp))
             SRL_iter_SET_OPTION(iter, SRL_F_DEDUPE_STRINGS); */
     }
-
-    return iter;
 }
 
 void
-srl_destroy_iterator(pTHX_ srl_iterator_t *iter)
+srl_copy_iterator(pTHX_ srl_iterator_t *from, srl_iterator_t *to)
 {
-    if (iter->stack) {
-        srl_stack_destroy(aTHX_ iter->stack);
-        Safefree(iter->stack);
-    }
+    srl_iterator_ptr iter = from;
+
+    assert(from != NULL);
+    assert(to != NULL);
+    SRL_ITER_TRACE("from=%p to=%p", from, to);
+
+    if (srl_stack_copy(aTHX_ &from->stack, &to->stack) != 0)
+        croak("Out of memory");
+
+    /* it's assumed that buf holds buffer owned by sv */
+    to->document = from->document;
+    if (to->document) SvREFCNT_inc(to->document);
+    Copy(&from->buf, &to->buf, 1, srl_reader_buffer_t);
+
+    to->pstack = &to->stack;
+    to->pbuf = &to->buf;
+    to->dec = NULL;
+
+    assert(to->buf.pos == from->buf.pos);
+}
+
+void
+srl_deinit_iterator(pTHX_ srl_iterator_t *iter)
+{
+    assert(iter != NULL);
 
     if (iter->dec)
         srl_destroy_decoder(aTHX_ iter->dec);
 
     if (iter->document)
         SvREFCNT_dec(iter->document);
+}
 
-    if (iter->tmp_buf_owner)
-        SvREFCNT_dec(iter->tmp_buf_owner);
-
+void
+srl_destroy_iterator(pTHX_ srl_iterator_t *iter)
+{
+    srl_deinit_iterator(iter);
     Safefree(iter);
-    return;
 }
 
 void
 srl_iterator_set(pTHX_ srl_iterator_t *iter, SV *src)
 {
+    SV *sv;
     STRLEN len;
     UV header_len;
     U8 encoding_flags;
     U8 protocol_version;
     srl_reader_char_ptr tmp;
     IV proto_version_and_encoding_flags_int;
-
-    /* clear previous buffer */
-    if (iter->tmp_buf_owner) {
-        SvREFCNT_dec(iter->tmp_buf_owner);
-        iter->tmp_buf_owner = NULL;
-    }
+    srl_iterator_stack_ptr stack_ptr = NULL;
 
     if (iter->document) {
         SvREFCNT_dec(iter->document);
@@ -292,11 +301,15 @@ srl_iterator_set(pTHX_ srl_iterator_t *iter, SV *src)
     } else if (   encoding_flags == SRL_PROTOCOL_ENCODING_SNAPPY
                || encoding_flags == SRL_PROTOCOL_ENCODING_SNAPPY_INCREMENTAL)
     {
-        srl_decompress_body_snappy(aTHX_ iter->pbuf, encoding_flags, &iter->tmp_buf_owner);
-        SvREFCNT_inc(iter->tmp_buf_owner);
+        srl_decompress_body_snappy(aTHX_ iter->pbuf, encoding_flags, &sv);
+        SvREFCNT_dec(iter->document);
+        SvREFCNT_inc(sv);
+        iter->document = sv;
     } else if (encoding_flags == SRL_PROTOCOL_ENCODING_ZLIB) {
-        srl_decompress_body_zlib(aTHX_ iter->pbuf, &iter->tmp_buf_owner);
-        SvREFCNT_inc(iter->tmp_buf_owner);
+        srl_decompress_body_zlib(aTHX_ iter->pbuf, &sv);
+        SvREFCNT_dec(iter->document);
+        SvREFCNT_inc(sv);
+        iter->document = sv;
     } else {
         SRL_RDR_ERROR(iter->pbuf, "Sereal document encoded in an unknown format");
     }
@@ -305,14 +318,14 @@ srl_iterator_set(pTHX_ srl_iterator_t *iter, SV *src)
     SRL_RDR_UPDATE_BODY_POS(iter->pbuf, protocol_version);
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 
-    srl_stack_push_and_set(iter, SRL_ITER_STACK_ROOT_TAG, 1);
+    srl_stack_push_and_set(iter, SRL_ITER_STACK_ROOT_TAG, 1, stack_ptr);
     srl_iterator_reset(aTHX_ iter);
 }
 
 void
 srl_iterator_reset(pTHX_ srl_iterator_t *iter)
 {
-    srl_stack_t *stack = iter->stack;
+    srl_stack_t *stack = iter->pstack;
     SRL_ITER_TRACE("--------------------------");
 
     while (!SRL_ITER_STACK_ON_ROOT(stack)) {
@@ -326,7 +339,7 @@ void
 srl_iterator_unite(pTHX_ srl_iterator_t *iter)
 {
     UV offset;
-    srl_stack_t *stack = iter->stack;
+    srl_stack_t *stack = iter->pstack;
     SRL_ITER_TRACE("--------------------------");
 
     if (expect_false(SRL_STACK_DEPTH(stack) <= 0))
@@ -364,7 +377,7 @@ srl_iterator_disjoin(pTHX_ srl_iterator_t *iter)
     // current tag). By default stack keeps offset to tag's starting point
     // (i.e. continer located in the buf).
 
-    srl_stack_push_ptr(iter->stack, stack_ptr);
+    srl_stack_push_ptr(iter->pstack, stack_ptr);
     stack_ptr->offset = SRL_RDR_BODY_POS_OFS(iter->pbuf); // disjoint point
     stack_ptr->tag = SRL_ITER_STACK_ROOT_TAG;
     stack_ptr->count = 1;
@@ -375,11 +388,10 @@ srl_iterator_disjoin(pTHX_ srl_iterator_t *iter)
 SRL_STATIC_INLINE void
 srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter)
 {
-    srl_stack_t *stack = iter->stack;
     SRL_ITER_ASSERT_STACK(iter);
 
-    stack->ptr->idx = stack->ptr->count;
-    iter->buf.pos = iter->buf.body_pos + stack->ptr->offset;
+    iter->stack.ptr->idx = iter->stack.ptr->count;
+    iter->buf.pos = iter->buf.body_pos + iter->stack.ptr->offset;
 
     SRL_ITER_REPORT_STACK_STATE(iter);
     SRL_ITER_TRACE("ofs %"UVuf" body_ofs %"UVuf,
@@ -390,19 +402,19 @@ srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter)
 }
 
 #define srl_iterator_wrap_stack(iter, expected_depth, stack_ptr) STMT_START {                   \
-    srl_stack_t *stack = (iter)->stack;                                                         \
-    (stack_ptr) = stack->ptr;                                                                   \
-                                                                                                \
-    SRL_ITER_TRACE("expected_depth=%"IVdf, (expected_depth));                                   \
+    SRL_ITER_TRACE("expected_depth=%"IVdf, (IV) (expected_depth));                              \
     SRL_ITER_REPORT_STACK_STATE(iter);                                                          \
                                                                                                 \
-    assert(((expected_depth) > (IV) SRL_STACK_DEPTH(stack)) == 0);                              \
+    assert((((IV) (expected_depth)) > iter->stack.depth) == 0);                                 \
                                                                                                 \
-    while ((stack_ptr)->idx == 0 && (stack_ptr)->tag != SRL_ITER_STACK_ROOT_TAG) {              \
-        if (SRL_STACK_DEPTH(stack) == (expected_depth)) break;                                  \
-        srl_stack_pop_nocheck(stack);                                                           \
-        (stack_ptr) = stack->ptr;                                                               \
-        if (SRL_ITER_STACK_ON_ROOT(stack))                                                      \
+    while (    (iter)->stack.depth != (IV) (expected_depth)                                     \
+            && (stack_ptr)->idx == 0                                                            \
+            && (stack_ptr)->tag != SRL_ITER_STACK_ROOT_TAG                                      \
+          )                                                                                     \
+    {                                                                                           \
+        srl_stack_pop_nocheck((iter)->pstack);                                                  \
+        (stack_ptr) = (iter)->stack.ptr;                                                        \
+        if (SRL_ITER_STACK_ON_ROOT(iter->pstack))                                               \
             SRL_ITER_TRACE("root of stack reached");                                            \
     }                                                                                           \
 } STMT_END
@@ -414,7 +426,7 @@ srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter)
     UV length;                                                                                  \
                                                                                                 \
     if (expect_false((stack_ptr)->idx == 0))                                                    \
-        SRL_ITER_ERRORf1("Nothing to parse at depth=%"IVdf, SRL_STACK_DEPTH(iter->stack));      \
+        SRL_ITER_ERRORf1("Nothing to parse at depth=%"IVdf, SRL_STACK_DEPTH(iter->pstack));     \
                                                                                                 \
     (stack_ptr)->idx--;                                                                         \
                                                                                                 \
@@ -422,6 +434,10 @@ srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter)
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);                                                          \
                                                                                                 \
 read_again:                                                                                     \
+    if (expect_false(SRL_RDR_DONE(iter->pbuf))) {                                               \
+        SRL_ITER_ERROR("EOF is reached");                                                       \
+    }                                                                                           \
+                                                                                                \
     tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;                                                 \
     SRL_ITER_REPORT_TAG(iter, tag);                                                             \
     iter->buf.pos++;                                                                            \
@@ -429,80 +445,78 @@ read_again:
     /* No code which decrease step, next or stack's counters should be added here.              \
      * Otherwise the counters will be decreased twicer for tags like REFN, ALIAS, etc. */       \
                                                                                                 \
-    switch (tag) {                                                                              \
-        CASE_SRL_HDR_SHORT_BINARY:                                                              \
+    switch (tag & 0xE0) {                                                                       \
+        case 0x0: /* POS_0 .. NEG_1 */                                                          \
+            break;                                                                              \
+                                                                                                \
+        case 0x40: /* ARRAYREF_0 .. HASHREF_15 */                                               \
+            /* for HASHREF_0 .. HASHREF_15 multiple length by two */                            \
+            length = (tag & 0xF) << ((tag & 0x10) ? 1 : 0);                                     \
+            srl_stack_push_and_set(iter, tag, length, stack_ptr);                               \
+            break;                                                                              \
+                                                                                                \
+        case 0x60: /* SHORT_BINARY_0 .. SHORT_BINARY_31 */                                      \
             iter->buf.pos += SRL_HDR_SHORT_BINARY_LEN_FROM_TAG(tag);                            \
             break;                                                                              \
                                                                                                 \
-        case SRL_HDR_HASH:                                                                      \
-            length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading HASH");         \
-            if (length > 0) srl_stack_push_and_set(iter, tag, length * 2);                      \
-            break;                                                                              \
-                                                                                                \
-        case SRL_HDR_ARRAY:                                                                     \
-            length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading ARRAY");        \
-            if (length > 0) srl_stack_push_and_set(iter, tag, length);                          \
-            break;                                                                              \
-                                                                                                \
-        CASE_SRL_HDR_HASHREF:                                                                   \
-            length = SRL_HDR_HASHREF_LEN_FROM_TAG(tag);                                         \
-            if (length > 0) srl_stack_push_and_set(iter, tag, length * 2);                      \
-            break;                                                                              \
-                                                                                                \
-        CASE_SRL_HDR_ARRAYREF:                                                                  \
-            length = SRL_HDR_ARRAYREF_LEN_FROM_TAG(tag);                                        \
-            if (length > 0) srl_stack_push_and_set(iter, tag, length);                          \
-            break;                                                                              \
-                                                                                                \
-        CASE_SRL_HDR_POS:                                                                       \
-        CASE_SRL_HDR_NEG:                                                                       \
-            break;                                                                              \
-                                                                                                \
-        case SRL_HDR_VARINT:                                                                    \
-        case SRL_HDR_ZIGZAG:                                                                    \
-            srl_skip_varint(aTHX_ iter->pbuf);                                                  \
-            break;                                                                              \
-                                                                                                \
-        case SRL_HDR_FLOAT:         iter->buf.pos += 4;      break;                             \
-        case SRL_HDR_DOUBLE:        iter->buf.pos += 8;      break;                             \
-        case SRL_HDR_LONG_DOUBLE:   iter->buf.pos += 16;     break;                             \
-                                                                                                \
-        case SRL_HDR_TRUE:                                                                      \
-        case SRL_HDR_FALSE:                                                                     \
-        case SRL_HDR_UNDEF:                                                                     \
-        case SRL_HDR_CANONICAL_UNDEF:                                                           \
-            break;                                                                              \
-                                                                                                \
-        case SRL_HDR_REFN:                                                                      \
-        case SRL_HDR_ALIAS:                                                                     \
-        case SRL_HDR_WEAKEN:                                                                    \
-            goto read_again;                                                                    \
-                                                                                                \
-        case SRL_HDR_PAD:                                                                       \
-            while (SRL_RDR_NOT_DONE(iter->pbuf) && *iter->buf.pos++ == SRL_HDR_PAD) {};         \
-            goto read_again;                                                                    \
-                                                                                                \
-        case SRL_HDR_BINARY:                                                                    \
-        case SRL_HDR_STR_UTF8:                                                                  \
-            length = srl_read_varint_uv_length(aTHX_ iter->pbuf,                                \
-                                              " while reading BINARY or STR_UTF8");             \
-            iter->buf.pos += length;                                                            \
-            break;                                                                              \
-                                                                                                \
-        case SRL_HDR_COPY:                                                                      \
-        case SRL_HDR_REFP:                                                                      \
-            srl_skip_varint(aTHX_ iter->pbuf);                                                  \
-            break;                                                                              \
-                                                                                                \
-        /* case SRL_HDR_OBJECTV: */                                                             \
-        /* case SRL_HDR_OBJECTV_FREEZE: */                                                      \
-        /* case SRL_HDR_REGEXP: */                                                              \
-        /* case SRL_HDR_OBJECT: */                                                              \
-        /* case SRL_HDR_OBJECT_FREEZE: */                                                       \
-                                                                                                \
         default:                                                                                \
-            SRL_RDR_ERROR_UNIMPLEMENTED(iter->pbuf, tag, "");                                   \
-            break;                                                                              \
+            switch (tag) {                                                                      \
+                case SRL_HDR_HASH:                                                              \
+                    length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading HASH"); \
+                    srl_stack_push_and_set(iter, tag, length * 2, stack_ptr);                   \
+                    break;                                                                      \
+                                                                                                \
+                case SRL_HDR_ARRAY:                                                             \
+                    length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading ARRAY");\
+                    srl_stack_push_and_set(iter, tag, length, stack_ptr);                       \
+                    break;                                                                      \
+                                                                                                \
+                case SRL_HDR_VARINT:                                                            \
+                case SRL_HDR_ZIGZAG:                                                            \
+                    srl_skip_varint(aTHX_ iter->pbuf);                                          \
+                    break;                                                                      \
+                                                                                                \
+                case SRL_HDR_FLOAT:         iter->buf.pos += 4;      break;                     \
+                case SRL_HDR_DOUBLE:        iter->buf.pos += 8;      break;                     \
+                case SRL_HDR_LONG_DOUBLE:   iter->buf.pos += 16;     break;                     \
+                                                                                                \
+                case SRL_HDR_TRUE:                                                              \
+                case SRL_HDR_FALSE:                                                             \
+                case SRL_HDR_UNDEF:                                                             \
+                case SRL_HDR_CANONICAL_UNDEF:                                                   \
+                    break;                                                                      \
+                                                                                                \
+                case SRL_HDR_REFN:                                                              \
+                case SRL_HDR_ALIAS:                                                             \
+                case SRL_HDR_WEAKEN:                                                            \
+                    goto read_again;                                                            \
+                                                                                                \
+                case SRL_HDR_PAD:                                                               \
+                    while (SRL_RDR_NOT_DONE(iter->pbuf) && *iter->buf.pos++ == SRL_HDR_PAD) {}; \
+                    goto read_again;                                                            \
+                                                                                                \
+                case SRL_HDR_BINARY:                                                            \
+                case SRL_HDR_STR_UTF8:                                                          \
+                    length = srl_read_varint_uv_length(aTHX_ iter->pbuf,                        \
+                                                      " while reading BINARY or STR_UTF8");     \
+                    iter->buf.pos += length;                                                    \
+                    break;                                                                      \
+                                                                                                \
+                case SRL_HDR_COPY:                                                              \
+                case SRL_HDR_REFP:                                                              \
+                    srl_skip_varint(aTHX_ iter->pbuf);                                          \
+                    break;                                                                      \
+                                                                                                \
+                /* case SRL_HDR_OBJECTV: */                                                     \
+                /* case SRL_HDR_OBJECTV_FREEZE: */                                              \
+                /* case SRL_HDR_REGEXP: */                                                      \
+                /* case SRL_HDR_OBJECT: */                                                      \
+                /* case SRL_HDR_OBJECT_FREEZE: */                                               \
+                                                                                                \
+                default:                                                                        \
+                    SRL_RDR_ERROR_UNIMPLEMENTED(iter->pbuf, tag, "");                           \
+                    break;                                                                      \
+            }                                                                                   \
     }                                                                                           \
                                                                                                 \
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);                                                          \
@@ -513,7 +527,7 @@ read_again:
 void
 srl_iterator_step_in(pTHX_ srl_iterator_t *iter, UV n)
 {
-    srl_iterator_stack_ptr stack_ptr;
+    srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
@@ -536,9 +550,8 @@ srl_iterator_step_in(pTHX_ srl_iterator_t *iter, UV n)
 void
 srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
 {
-    srl_stack_t *stack = iter->stack;
-    srl_iterator_stack_ptr stack_ptr;
-    IV expected_depth = SRL_STACK_DEPTH(stack);
+    IV expected_depth = iter->stack.depth;
+    srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
@@ -551,7 +564,7 @@ srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
     while (1) {
         srl_iterator_wrap_stack(iter, expected_depth, stack_ptr);
 
-        if (SRL_STACK_DEPTH(stack) == expected_depth) {
+        if (iter->stack.depth == expected_depth) {
             if (n == 0) break;
             else n--;
         }
@@ -563,9 +576,9 @@ srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
         SRL_ITER_ERRORf1("Failed to do %"UVuf" next steps. Likely EOF was reached", n);
     }
 
-    if (expect_false(SRL_STACK_DEPTH(stack) != expected_depth)) {
+    if (expect_false(iter->stack.depth != expected_depth)) {
         SRL_ITER_ERRORf2("next() led to wrong stack depth, expected=%"IVdf", actual=%"IVdf,
-                          expected_depth, SRL_STACK_DEPTH(stack));
+                          expected_depth, iter->stack.depth);
     }
 
     SRL_ITER_TRACE("Did expected number of steps at depth %"IVdf, expected_depth);
@@ -577,9 +590,8 @@ srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
 
 void
 srl_iterator_next_until_depth_and_idx(pTHX_ srl_iterator_t *iter, UV expected_depth, U32 expected_idx) {
-    srl_stack_t *stack = iter->stack;
-    srl_iterator_stack_ptr stack_ptr = stack->ptr;
-    IV current_depth = SRL_STACK_DEPTH(stack);
+    IV current_depth = iter->stack.depth;
+    srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
@@ -596,7 +608,7 @@ srl_iterator_next_until_depth_and_idx(pTHX_ srl_iterator_t *iter, UV expected_de
                          expected_depth, current_depth);
     }
 
-    stack_ptr = stack->begin + expected_depth;
+    stack_ptr = iter->stack.begin + expected_depth;
     if (expect_false((IV) expected_idx > stack_ptr->idx)) {
         SRL_ITER_ERRORf3("srl_iterator_next_until_depth() can only go forward, "
                          "so expected_idx=%u should not be greater then current "
@@ -604,12 +616,12 @@ srl_iterator_next_until_depth_and_idx(pTHX_ srl_iterator_t *iter, UV expected_de
                          expected_idx, stack_ptr->idx, expected_depth);
     }
 
-    stack_ptr = stack->ptr;
+    stack_ptr = iter->stack.ptr;
 
     while (1) {
         srl_iterator_wrap_stack(iter, expected_depth, stack_ptr);
 
-        if (SRL_STACK_DEPTH(stack) == (IV) expected_depth) {
+        if (iter->stack.depth == (IV) expected_depth) {
             if (stack_ptr->idx == (IV) expected_idx)  break;
             assert(((IV) expected_idx > stack_ptr->idx) == 0);
         }
@@ -617,8 +629,8 @@ srl_iterator_next_until_depth_and_idx(pTHX_ srl_iterator_t *iter, UV expected_de
         srl_iterator_step_internal(iter, stack_ptr);
     }
 
-    assert(stack_ptr->idx == expected_idx);
-    assert(SRL_STACK_DEPTH(stack) == expected_depth);
+    assert(stack_ptr->idx == (IV) expected_idx);
+    assert(iter->stack.depth == (IV) expected_depth);
     SRL_ITER_TRACE("Reached expected stack depth: %"UVuf " and idx: %u",
                    expected_depth, expected_idx);
 
@@ -628,7 +640,7 @@ srl_iterator_next_until_depth_and_idx(pTHX_ srl_iterator_t *iter, UV expected_de
 void
 srl_iterator_step_out(pTHX_ srl_iterator_t *iter, UV n)
 {
-    srl_stack_t *stack = iter->stack;
+    srl_stack_t *stack = iter->pstack;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
@@ -654,8 +666,7 @@ IV
 srl_iterator_array_goto(pTHX_ srl_iterator_t *iter, I32 idx)
 {
     I32 s_idx;
-    srl_stack_t *stack = iter->stack;
-    srl_iterator_stack_ptr stack_ptr = stack->ptr;
+    srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_EOF(iter, "array element");
@@ -789,9 +800,8 @@ srl_iterator_hash_exists(pTHX_ srl_iterator_t *iter, const char *name, STRLEN na
     UV length, offset;
     const char *key_ptr;
 
-    srl_stack_t *stack = iter->stack;
-    IV stack_depth = SRL_STACK_DEPTH(stack);
-    srl_iterator_stack_ptr stack_ptr = stack->ptr;
+    IV stack_depth = iter->stack.depth;
+    srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_EOF(iter, "stringish");
@@ -805,7 +815,7 @@ srl_iterator_hash_exists(pTHX_ srl_iterator_t *iter, const char *name, STRLEN na
         stack_ptr->idx--; // do not make it be part of while clause
         SRL_ITER_ASSERT_STACK(iter);
         assert(stack_ptr->idx % 2 == 1);
-        assert(SRL_STACK_DEPTH(stack) == stack_depth);
+        assert(iter->stack.depth == stack_depth);
         DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 
         tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;
@@ -879,7 +889,7 @@ srl_iterator_hash_exists(pTHX_ srl_iterator_t *iter, const char *name, STRLEN na
 
         // srl_iterator_next garantee that we remans on current stack
         srl_iterator_next(aTHX_ iter, 1);
-        stack_ptr = stack->ptr;
+        stack_ptr = iter->stack.ptr;
     }
 
     SRL_ITER_TRACE("didn't found key '%.*s'", (int) name_len, name);
@@ -975,28 +985,28 @@ read_again:
 srl_iterator_stack_ptr
 srl_iterator_stack(pTHX_ srl_iterator_t *iter)
 {
-    return srl_stack_empty(iter->stack) ? NULL : iter->stack->ptr;
+    return srl_stack_empty(iter->pstack) ? NULL : iter->stack.ptr;
 }
 
 IV
 srl_iterator_stack_depth(pTHX_ srl_iterator_t *iter)
 {
-    return SRL_STACK_DEPTH(iter->stack);
+    return SRL_STACK_DEPTH(iter->pstack);
 }
 
 UV
 srl_iterator_stack_index(pTHX_ srl_iterator_t *iter)
 {
     SRL_ITER_ASSERT_STACK(iter);
-    assert((I32) iter->stack->ptr->count >= iter->stack->ptr->idx); 
-    return (UV) (iter->stack->ptr->count - iter->stack->ptr->idx);
+    assert((I32) iter->stack.ptr->count >= iter->stack.ptr->idx); 
+    return (UV) (iter->stack.ptr->count - iter->stack.ptr->idx);
 }
 
 UV
 srl_iterator_stack_info(pTHX_ srl_iterator_t *iter, UV *length_ptr)
 {
     UV type = 0;
-    srl_stack_t *stack = iter->stack;
+    srl_stack_t *stack = iter->pstack;
     srl_iterator_stack_ptr stack_ptr = stack->ptr;
 
     SRL_ITER_ASSERT_STACK(iter);
